// $Id: PDFStyle.java 10958 2009-09-29 13:19:54Z chris $

package org.faceless.pdf;

import java.util.*;
import java.awt.Color;
import java.awt.Paint;
import java.awt.GradientPaint;

/**
 * <p>
 * A PDFStyle controls the colors, font and many other aspects of the actual
 * display of elements on a PDF page. It's conceptually similar to a CSS
 * style used with HTML markup.
 * </p>
 * <p>
 * The idea behind the PDFStyle class is that you create a style and then
 * apply it to the page. This means you can switch from one style to another
 * with a single command, and switch back just as easily. It makes defining
 * a consistant "feel" to your document much easier than it would if you had
 * to set the font and color separately.
 * </p>
 * <b>Example</b>
 * <pre>
 *   import java.awt.Color;
 *
 *   // Create a new style "normal": 12pt black Times-Roman, with
 *   // line-spacing of 1.5
 *   PDFStyle normal = new PDFStyle();
 *   normal.setFont(new StandardFont(StandardFont.TIMES), 12);
 *   normal.setFillColor(Color.black);
 *   normal.setTextLineSpacing(1.5);
 *
 *   // Create a new varient of "normal": 12pt red Times-Italic with
 *   // the same line-spacing
 *   PDFStyle italic = (PDFStyle)normal.clone();
 *   normal.setFont(new StandardFont(StandardFont.TIMESITALIC), 12);
 *   normal.setFillColor(Color.red);
 *
 *   // Create a style to draw a box around the text in green, with a
 *   // line width of 2 points
 *   PDFStyle boxstyle = new PDFStyle();
 *   boxstyle.setLineColor(Color.green);
 *   boxstyle.setLineWeighting(2);
 *
 *   // Now use these styles in a document
 *   PDF p = new PDF();
 *   PDFPage page = p.newPage(PDF.PAGESIZE_A4);
 *   page.setStyle(normal);
 *   page.drawText("This is in 12pt black Times Roman", 100, 100);
 *   page.setStyle(italic);
 *   page.drawText("This is in 12pt red Times Italic", 100, 120);
 *
 *   // Draw a box around them in green
 *   page.setStyle(boxstyle);
 *   page.drawRectangle(90,90, 200, 40);
 * </pre>
 *
 * <p>
 * This shows several useful aspects of styles:
 * <ul>
 * <li>Styles can easily be extended by using the <code>clone()</code> method.</li>
 * <li>Not all the aspects of a style need to be set.</li>
 * <li>Styles can be given meaningful names.</li>
 * </ul>
 * </p>
 * @version $Revision: 1.6 $
 */
public class PDFStyle extends PeeredObject implements Cloneable
{
    /**
     * Set the text alignment for this style to left-aligned
     */
    public static final int TEXTALIGN_LEFT = 1;

    /**
     * Set the text alignment for this style to right-aligned
     */
    public static final int TEXTALIGN_RIGHT = 2;

    /**
     * Set the text alignment for this style to centered
     */
    public static final int TEXTALIGN_CENTER = 4;

    /**
     * Set the text alignment for this style to justified (the default).
     * Justification is only useful when text is written within a paragraph.
     * When a single element of text is drawn, the effect is the same as
     * left-alignment.
     */
    public static final int TEXTALIGN_JUSTIFY = 8;

    /**
     * Set the vertical text alignment for this style to baseline (the default).
     */
    public static final int TEXTALIGN_BASELINE = 0;

    /**
     * Set the vertical text alignment for this style to top
     */
    public static final int TEXTALIGN_TOP = 16;

    /**
     * Set the vertical text alignment for this style to middle
     */
    public static final int TEXTALIGN_MIDDLE = 32;

    /**
     * Set the vertical text alignment for this style to bottom
     */
    public static final int TEXTALIGN_BOTTOM = 64;

    /**
     * Set the end of a line to be squared off at the end. There is no
     * projection beyond the end of the path. This is the default
     */
    public static final int LINECAP_BUTT = 0;

    /**
     * Set the end of a line to be rounded at the end. Effectively draws a circle with
     * a diameter of the line width at the end of each line.
     */
    public static final int LINECAP_ROUND = 1;

    /**
     * Set the end of a line to be squared at the end. Effectively extends each line by
     * half the linewidth.
     */
    public static final int LINECAP_SQUARE = 2;

    /**
     * Sets the join style of two lines so that the lines are extended so they meet at
     * a point (like a picture frame). For extremely sharp angles, this will automatically
     * be converted to a LINEJOIN_BEVEL. This is the default.
     */
    public static final int LINEJOIN_MITER = 0;

    /**
     * Sets the join style of two lines so that the lines are rounded, equivalent to
     * drawing a circle with a diameter of the linewidth where the lines meet.
     */
    public static final int LINEJOIN_ROUND = 1;

    /**
     * Sets the join style of two lines so that the lines are beveled. The two lines
     * are drawn with LINECAP_BUTT ends, and the notch between the two segments is
     * filled in with a triangle.
     */
    public static final int LINEJOIN_BEVEL = 2;

    /**
     * Set any text rendered in this style to be filled with the styles
     * FillColor (the default)
     */
    public static final int FONTSTYLE_FILLED = 0;

    /**
     * Set any text rendered in this style to be drawn as a hollow outline
     * with the styles LineColor (the default)
     */
    public static final int FONTSTYLE_OUTLINE = 1;

    /**
     * Set any text rendered in this style to be filled with the styles
     * FillColor, then to be outlined with the styles' LineColor
     */
    public static final int FONTSTYLE_FILLEDOUTLINE = 2;

    /**
     * Set any text rendered in this style to be invisible. This becomes
     * useful in applications like OCR, where the original scanned image
     * is displayed on the screen and invisible text written above it,
     * allowing the text to be cut and pasted.
     */
    public static final int FONTSTYLE_INVISIBLE = 3;

    /**
     * A parameter to {@link #setPaintMethod} to set the paint method to
     * use the non-zero winding number method to determine which areas are
     * inside or outside a shape. This is the default.
     */
    public static final int PAINTMETHOD_NONZEROWINDING = 0;

    /**
     * A parameter to {@link #setPaintMethod} to set the paint method to
     * use the even-odd method to determine which areas are inside or
     * outside a shape.
     */
    public static final int PAINTMETHOD_EVENODD = 1;


    final org.faceless.pdf2.PDFStyle style;

    Object getPeer()
    {
        return style;
    }

    PDFStyle(org.faceless.pdf2.PDFStyle style)
    {
        this.style=style;
    }

    /**
     * A PDFStyle which can be passed in to {@link PDFPage#beginTextLink}
     * to underline the text links.
     */
    public static final PDFStyle LINKSTYLE;

    static {
        LINKSTYLE = new PDFStyle();
	LINKSTYLE.setTextUnderline(true);
    }

    /**
     * Create a new PDFStyle. The default aspects of this style are:
     * <ul>
     * <li>No Font specified, no line color, no fill color. These must be set by the user</li>
     * <li>Line Weighting of 1 point</li>
     * <li>Line Dashing set to a solid line</li>
     * <li>Line Spacing set to a 1</li>
     * <li>Text Justification ratio of 0.5</li>
     * <li>Line ends set to {@link #LINECAP_BUTT}</li>
     * <li>Line joints set to {@link #LINEJOIN_MITER}</li>
     * <li>Font style set to {@link #FONTSTYLE_FILLED}</li>
     * <li>Text alignment is locale dependent - for hebrew, arabic, urdu and yiddish, it's {@link #TEXTALIGN_RIGHT}. For japanese, chinese and korean, it's {@link #TEXTALIGN_LEFT} and for everything else it's {@link #TEXTALIGN_JUSTIFY}.</li>
     * </ul>
     */
    public PDFStyle()
    {
	style = new org.faceless.pdf2.PDFStyle();
    }

    /**
     * Set the text justification ratio for a style. When a line of text
     * is padded to justify it against two margins, there is always the
     * the option of extending the space between each character, extending
     * the space between each word, or a combination of the two. The
     * "justification ratio" determines how much to apply to each. A ratio
     * of 0 means "only extend the spaces between words", a ratio of 1
     * means "only extend the spaces between letters". The default is 0.5.
     */
    public void setTextJustificationRatio(float in)
    {
	style.setTextJustificationRatio(in);
    }

    /**
     * Set whether text rendered with this style is underlined or not.
     * The exact position and width of the underlining is font specific.
     * The default is <i>false</i>.
     * @since 1.1
     */
    public void setTextUnderline(boolean on)
    {
	style.setTextUnderline(on);
    }

    /**
     * Set whether text rendered with this style is struck-out or not.
     * The default is <i>false</i>.
     * @since 1.1
     */
    public void setTextStrikeOut(boolean on)
    {
	style.setTextStrikeOut(on);
    }

    /**
     * Set the line cap style. Value can be {@link #LINECAP_BUTT}, {@link #LINECAP_ROUND} or
     * {@link #LINECAP_SQUARE}
     * <p>
     * Changes to this setting midway through a <code>path</code> don't
     * take effect until it is closed.
     * </p>
     */
    public void setLineCap(int type)
    {
	style.setLineCap(type);
    }

    /**
     * Set the line join style. Value can be {@link #LINEJOIN_MITER}, {@link #LINEJOIN_ROUND} or
     * {@link #LINEJOIN_BEVEL}
     * <p>
     * Changes to this setting midway through a <code>path</code> don't
     * take effect until it is closed.
     * </p>
     */
    public void setLineJoin(int type)
    {
	style.setLineJoin(type);
    }

    /**
     * <p>
     * Set the line dashing pattern. The first <code>on</code> points of
     * the line will be drawn, then the next <code>off</code> points of
     * the line will be skipped. The <code>phase</code> argument defines
     * how far into this pattern to start.
     * </p>
     * <p>
     * An example would be <code>setLineDash(2,2,0)</code> to draw lines that
     * have two points on, two points off and so on, starting with two "on"
     * points. For solid lines, the call should
     * be <code>setLineDash(0,0,0)</code>
     * </p>
     * <p>
     * Changes to this setting midway through a <code>path</code> don't
     * take effect until it is closed.
     * </p>
     * Since 1.1.5, the parameters are floats rather than ints
     * @param on how many points of the line to draw
     * @param off how many points of the line to skip
     * @param phase how far into the pattern to start
     */
    public void setLineDash(float on, float off, float phase)
    {
	style.setLineDash(on,off,phase);
    }

    /**
     * Set the line weighting, for fonts and geometric shapes drawn as
     * outlines. The minimum possible value is zero, which indicates to
     * the PDF renderer to use the thinnest line possible. On high-resolution
     * devices, this will become nearly invisible. Because of its
     * device-dependent nature, it's not recommended.
     * <p>
     * Changes to this setting midway through a <code>path</code> don't
     * take effect until it is closed.
     * </p>
     *
     * @param thickness The thickness of the line, in points
     */
    public void setLineWeighting(float thickness)
    {
	style.setLineWeighting(thickness);
    }

    /**
     * <p>
     * Set the spacing between lines of text. In versions prior to 1.1,
     * this was a multiplied by the the fonts point size to give the difference
     * between the baselines of two successive lines of text, and the default
     * value was 1.2.
     * </p><p>
     * Since 1.1, this has been changed to a ratio of the <i>default spacing
     * between lines for this font</i> (as determined by the
     * {@link PDFFont#getDefaultLeading} method). This allows individual fonts to set
     * their preferred line spacing more accurately.
     * Although the default setting should result in nearly identical results
     * for most fonts, developers calling this method who are upgrading from
     * versions 1.0.4 or earlier should divide the value passed in by 1.2 to
     * meet the new method contract.
     * </p><p>
     * The default value is now 1. A value of 1.5 sets line-and-a-half spacing,
     * a value of 2 gives double spacing, and so on.
     * </p>
     * @param factor the spacing between lines
     */
    public void setTextLineSpacing(float factor)
    {
	style.setTextLineSpacing(factor);
    }

    /**
     * Set the text-alignment. Can be one of {@link #TEXTALIGN_LEFT},
     * {@link #TEXTALIGN_RIGHT}, {@link #TEXTALIGN_CENTER} or {@link #TEXTALIGN_JUSTIFY}.
     */
    public void setTextAlign(int align)
    {
	style.setTextAlign(align);
    }

    /**
     * <p>
     * Set the text vertical offset - the distance between the standard
     * baseline and the basline for this style, as a proportion of the
     * font size. This is mainly used for superscripting or subscripting.
     * text.
     * </p><p>
     * As an example, if you wanted to create a line of text that was 6 points
     * high and 6 points above the standard baseline, set the font size to
     * 6 and the TextRise() to 1. (1x6 = 6 points above the baseline).
     * To place the same text 3 points below the baseline, set the offset to -0.5.
     * </p>
     * @since 1.1
     */
    public void setTextRise(float offset)
    {
	style.setTextRise(offset);
    }

    /**
     * <p>
     * Set the fill color. For text, this is the color of the text.
     * For geometric shapes, this is paint to fill those shapes with.
     * To draw only outlines, set it to <code>null</code> (the default).
     * </p><p>
     * Prior to release 1.2, this method took a
     * {@link java.awt.Color} as an argument, but this has been changed to
     * its superclass, {@link java.awt.Paint}, instead.
     * Although most of the time the parameter will still be a plain color,
     * this change allows a {@link java.awt.GradientPaint} to be used
     * as well.
     * </p>
     * @param paint the paint to use, or <code>null</code> if no fill
     * is required.
     */
    public void setFillColor(Paint paint)
    {
	if (paint!=null && paint instanceof ColorPattern) {
	    paint = ((ColorPattern)paint).pattern;
	}
	style.setFillColor(paint);
    }

    /**
     * Set the line color. For text, this is the color of the outline of
     * the text. For geometric shapes, this is color to draw the outlines
     * of those shapes. If no outline is required, set it to
     * <code>null</code> (the default).
     * @param color The Color to use, or <code>null</code> if no outline
     * is required.
     */
    public void setLineColor(Color color)
    {
	style.setLineColor(color);
    }

    /**
     * <p>
     * Set the paint method to either {@link #PAINTMETHOD_EVENODD} or
     * {@link #PAINTMETHOD_NONZEROWINDING} (the default). The paint method
     * determines which area of a self-intersecting polygon is filled. If
     * your polygons aren't self-intersecting, it has no effect.
     * </p><p>
     * This setting is very obscure and is really only here for completeness.
     * For a full discussion of the difference see the <a href="http://partners.adobe.com/asn/developer/acrosdk/docs/PDFRef.pdf">PDF Reference manual version 1.3</a>, page 156.
     * </p>
     * @since 1.1.5
     */
    public void setPaintMethod(int method)
    {
	int newmethod = style.PAINTMETHOD_EVENODD;
	if (method==PAINTMETHOD_NONZEROWINDING) newmethod = style.PAINTMETHOD_NONZEROWINDING;
	style.setPaintMethod(newmethod);
    }

    /**
     * Set the font and font size for the style.
     * @param font the <code>PDFFont</code> to use
     * @param size the size of the font, in points
     */
    public void setFont(PDFFont font, float size)
    {
	style.setFont(font==null ? null : font.font, size);
    }

    /**
     * Add a backup font to the current style. Backup fonts are used when
     * the standard font doesn't have a character defined - for example, creating
     * a style with the "Times Roman" font and then adding a backup font of
     * "Symbol" would allow text to be written in both English and Greek without
     * changing styles. As many backup fonts as are required may be added.
     * @param font the font to add as a backup for the current style.
     * @since 1.2
     */
    public void addBackupFont(PDFFont font)
    {
	style.addBackupFont(font==null ? null : font.font);
    }

    /**
     * Get the specified backup font. The argument to this method determines
     * which backup font to return - zero for the first, one for the second and
     * so on.
     * @param i the backup font to return
     * @return the specified backup font, or <code>null</code> if no backup font
     * exists at that index.
     * @since 1.2
     */
    public PDFFont getBackupFont(int i)
    {
	return (PDFFont)PeeredObject.getPeer(style.getBackupFont(i));
    }

    /**
     * Set the font render style. Can be one of {@link #FONTSTYLE_FILLED}
     * (the default), {@link #FONTSTYLE_OUTLINE},
     * {@link #FONTSTYLE_FILLEDOUTLINE} or {@link #FONTSTYLE_INVISIBLE}.
     * @since 1.1
     */
    public void setFontStyle(int fontstyle)
    {
	int newfontstyle=org.faceless.pdf2.PDFStyle.FONTSTYLE_FILLED;
	if (fontstyle==FONTSTYLE_OUTLINE) newfontstyle=org.faceless.pdf2.PDFStyle.FONTSTYLE_OUTLINE;
	else if (fontstyle==FONTSTYLE_FILLEDOUTLINE) newfontstyle=org.faceless.pdf2.PDFStyle.FONTSTYLE_FILLEDOUTLINE;
	else if (fontstyle==FONTSTYLE_INVISIBLE) newfontstyle=org.faceless.pdf2.PDFStyle.FONTSTYLE_INVISIBLE;
	style.setFontStyle(newfontstyle);
    }

    /**
     * <p>
     * Set the number of points to indent the first line of any text
     * drawn in this style. Positive values result in the first line
     * being indented to the right, negative values in the first line
     * being indented to the left (this is reversed for RTL scripts)
     * </p><p>
     * Note that mixing styles with different text-indent levels on
     * the first line of text in a paragraph will result in
     * unpredictable results.
     * </p>
     * @since 1.1.21
     */
    public void setTextIndent(float indent)
    {
	style.setTextIndent(indent);
    }

    /**
     * Get the length of the specified string in points, using the
     * styles font and font size. The line of text must not contain
     * any newlines. For horizontal fonts, this is equivalent to
     * <code>getTextRight()-getTextLeft()</code>.
     * This method takes track-kerning into account.
     */
    public float getTextLength(String s)
    {
	return style.getTextLength(s);
    }

    /**
     * Get the length of the specified string in points, using the
     * styles font and font size. As for {@link #getTextLength(String)}
     * but takes a char array, and the text must be {@link PDFFont#ligaturize ligaturized}
     * first if necessary.
     * @since 1.2.1
     */
    public float getTextLength(char[] c, int off, int len)
    {
	return style.getTextLength(c,off,len);
    }

    /**
     * Get the left-most X co-ordinate of the specified string in points if it
     * was rendered at (0,0), using the styles font and font size. The line of
     * text must not contain any newlines.
     * This method takes track-kerning into account.
     */
    public float getTextLeft(String s)
    {
	return style.getTextLeft(s);
    }

    /**
     * Get the right-most X co-ordinate of the specified string in points if it
     * was rendered at (0,0), using the styles font and font size. The line of
     * text must not contain any newlines.
     * This method takes track-kerning into account.
     */
    public float getTextRight(String s)
    {
	return style.getTextRight(s);
    }

    /**
     * Get the top-most Y co-ordinate of the specified string in points if it
     * was rendered at (0,0), using the styles font and font size. The line of
     * text must not contain any newlines.
     * This method takes track-kerning into account.
     */
    public float getTextTop(String s)
    {
	return style.getTextTop(s);
    }

    /**
     * Get the bottom-most Y co-ordinate of the specified string in points if it
     * was rendered at (0,0), using the styles font and font size. The line of
     * text must not contain any newlines.
     * This method takes track-kerning into account.
     */
    public float getTextBottom(String s)
    {
	return style.getTextBottom(s);
    }

    /**
     * Get the text leading for this styles in points. The leading is the
     * distance between the baseline of two successive lines of text, and
     * is defined as:<BR>
     * <code>getFontSize() * font.getDefaultLeading() * getLineSpacing()</code>
     */
    public float getFontLeading()
    {
	return style.getFontLeading();
    }

    /**
     * Get the styles line thickness, in points.
     */
    public float getLineWeighting()
    {
	return style.getLineWeighting();
    }

    /**
     * Get the paint method as set by {@link #setPaintMethod}
     */
    public int getPaintMethod()
    {
	return PAINTMETHOD_EVENODD;
    }

    /**
     * Get the spacing between lines of text, as a ratio of default spacing
     * between lines of this point size - i.e. the value is "1" for single
     * spaced, "2" for double spaced and so on.
     */
    public float getTextLineSpacing()
    {
	return style.getTextLineSpacing();
    }

    /**
     * Get the styles text justification ratio
     */
    public float getTextJustificationRatio()
    {
	return 0.5f;
    }

    /**
     * Get whether the style has text underlining set
     */
    public boolean getTextUnderline()
    {
	return false;
    }

    /**
     * Get whether the style has text strikeout set
     */
    public boolean getTextStrikeOut()
    {
	return false;
    }

    /**
     * Get the styles' fill color. Since 1.2, this method returns a
     * {@link java.awt.Paint} instead of its subclass, {@link java.awt.Color}.
     */
    public Paint getFillColor()
    {
	return style.getFillColor();
    }

    /**
     * Get the styles line color
     */
    public Color getLineColor()
    {
	return (Color)style.getLineColor();
    }

    /**
     * Get the styles font
     */
    public PDFFont getFont()
    {
	return (PDFFont)PeeredObject.getPeer(style.getFont());
    }

    /**
     * Get the styles font size in points.
     */
    public float getFontSize()
    {
	return style.getFontSize();
    }

    /**
     * Get the styles line join style.
     */
    public int getLineJoin()
    {
	return style.getLineJoin();
    }

    /**
     * Get the styles line cap style.
     */
    public int getLineCap()
    {
	return style.getLineCap();
    }

    /**
     * Get the styles line dash pattern. The returned array is two elements
     * long, with element zero being the "on" and element one the "off".
     * Since 1.1.5, this value is a float (rather than int)
     */
    public float[] getLineDashPattern()
    {
	float[] f = new float[2];
	f[0] = style.getLineDashOn();
	f[1] = style.getLineDashOff();
	return f;
    }

    /**
     * Get the styles line dash phase - how far into the dash pattern the
     * line should start. Since 1.1.5, this value is a float (rather than int)
     */
    public float getLineDashPhase()
    {
	return style.getLineDashPhase();
    }

    /**
     * Get the styles text alignment
     */
    public int getTextAlign()
    {
        return 0;
    }

    /**
     * Get the font style.
     */
    public int getFontStyle()
    {
	int newfontstyle=style.getFontStyle();
	if (newfontstyle==org.faceless.pdf2.PDFStyle.FONTSTYLE_OUTLINE) return FONTSTYLE_OUTLINE;
	else if (newfontstyle==org.faceless.pdf2.PDFStyle.FONTSTYLE_FILLEDOUTLINE) return FONTSTYLE_FILLEDOUTLINE;
	else if (newfontstyle==org.faceless.pdf2.PDFStyle.FONTSTYLE_INVISIBLE) return FONTSTYLE_INVISIBLE;
	return FONTSTYLE_FILLED;
    }

    /**
     * Get the text vertical offset for this style
     */
    public float getTextRise()
    {
	return 0;
    }

    /**
     * Get the track kerning value as set by {@link #setTrackKerning}
     * @since 1.1.14
     */
    public float getTrackKern()
    {
        return 0;
    }

    /**
     * Get the text indent value as set by {@link #setTextIndent}
     * @since 1.1.21
     */
    public float getTextIndent()
    {
        return style.getTextIndent();
    }

    /**
     * Get a shallow copy of this style.
     */
    public Object clone()
    {
	return PeeredObject.getPeer(style.clone());
    }

    /**
     * Convenience Method - Get a clone of this style, typecast to PDFStyle
     */
    public PDFStyle styleClone()
    {
        return (PDFStyle)clone();
    }

    /**
     * Return a new style which is the "superscripted" version of
     * the current style.
     * @since 1.1
     */
    public PDFStyle superscriptClone()
    {
	return new PDFStyle(style.superscriptClone());
    }

    /**
     * Return a new style which is the "subscripted" version of
     * the current style.
     * @since 1.1
     */
    public PDFStyle subscriptClone()
    {
	return new PDFStyle(style.subscriptClone());
    }

    /**
     * <p>
     * Allows you to explicitly set the kerning between characters for a font.
     * This method may be called as many times as necessary - between each
     * character if required - to set the kerning distance between characters
     * for all future text rendered in this style ("characters" in this context
     * includes spaces). This "track kerning" is used as well as the standard
     * "pair-wise" kerning, as returned by {@link PDFSimpleFont#getKerning}.
     * </p><p>
     * If text-alignment is set to {@link #TEXTALIGN_JUSTIFY}, kerning (both
     * track and pairwise) is scaled up or down depending on the justification
     * required.
     * </p>
     * @param millipoints the space to place between each character in millipoints
     * (thousandths of a point) if this font was rendered one point high. May be
     * positive, which moves the characters apart, or negative to move them closer
     * together.
     * @see PDFSimpleFont#getKerning
     *
     * @since 1.1.14
     */
    public void setTrackKerning(float millipoints)
    {
	style.setTrackKerning(millipoints);
    }

    /**
     * Style for {@link #setFormStyle} which draws a solid border
     * around the field (the default)
     */
    public static final int FORMSTYLE_SOLID=0;

    /**
     * Style for {@link #setFormStyle} which draws an border around the
     * field so that it looks inset into the page.
     */
    public static final int FORMSTYLE_INSET=1;

    /**
     * Style for {@link #setFormStyle} which draws an border around the
     * field so that it looks beveled.
     */
    public static final int FORMSTYLE_BEVEL=2;

    /**
     * Style for {@link #setFormStyle} which draws a single line under the
     * field
     */
    public static final int FORMSTYLE_UNDERLINE=3;

    /**
     * Sets the style of a form fields background to one of {@link #FORMSTYLE_SOLID},
     * {@link #FORMSTYLE_INSET}, {@link #FORMSTYLE_BEVEL} or
     * {@link #FORMSTYLE_UNDERLINE}. This method can be applied to the background
     * style passed into the <code>setStyle</code> method of a {@link FormElement},
     * but for all other purposes such as setting the style of a page, this setting is
     * ignored.
     * @param style the type of background to draw the form field.
     * @since 1.1.23
     */
    public void setFormStyle(int formstyle)
    {
	int newstyle=style.FORMSTYLE_SOLID;
	if (formstyle==FORMSTYLE_SOLID)			newstyle=style.FORMSTYLE_SOLID;
	else if (formstyle==FORMSTYLE_BEVEL)		newstyle=style.FORMSTYLE_BEVEL;
	else if (formstyle==FORMSTYLE_INSET)		newstyle=style.FORMSTYLE_INSET;
	else if (formstyle==FORMSTYLE_UNDERLINE)	newstyle=style.FORMSTYLE_UNDERLINE;
	style.setFormStyle(newstyle);
    }

    /**
     * Returns the form-style of the current style, as set by {@link #setFormStyle}.
     * @since 1.1.23
     */
    public int getFormStyle()
    {
	int formstyle = style.getFormStyle();
	int newstyle=FORMSTYLE_SOLID;
	if (formstyle==style.FORMSTYLE_SOLID)		newstyle=FORMSTYLE_SOLID;
	else if (formstyle==style.FORMSTYLE_BEVEL)	newstyle=FORMSTYLE_BEVEL;
	else if (formstyle==style.FORMSTYLE_INSET)	newstyle=FORMSTYLE_INSET;
	else if (formstyle==style.FORMSTYLE_UNDERLINE)	newstyle=FORMSTYLE_UNDERLINE;

	return newstyle;
    }
}
